/* global browser, Port, buildNodes, mediaType */

'use strict';

var init = function(win, doc, head) {
	if (typeof Port === 'undefined' || !doc || !doc.body || !head || !head.prefs) {
		return;
	}

	var cfg = JSON.parse(head.prefs);
	var media = doc.body.querySelector('img' + (cfg.video ? ', video' : ''));

	if (!media) {
		init = null;
		return;
	}

	var pdsp = function(e, d, p) {
		if (!e || !e.preventDefault || !e.stopPropagation) {
			return;
		}

		if (d === void 0 || d === true) {
			e.preventDefault();
		}

		if (p === false) {
			return;
		}

		e.stopImmediatePropagation();
	};

	var shortcut = {
		specKeys: {
			8: 'Backspace', 9: 'Tab', 13: 'Enter',
			16: 'shift', 17: 'ctrl', 18: 'alt', // control keys are not capital
			27: 'Esc', 32: 'Space',
			33: 'PgUp', 34: 'PgDn', 35: 'End', 36: 'Home',
			37: 'Left', 38: 'Up', 39: 'Right', 40: 'Down',
			45: 'Ins', 46: 'Del',
			96: '0', 97: '1', 98: '2', 99: '3', 100: '4',
			101: '5', 102: '6', 103: '7', 104: '8', 105: '9',
			106: '*', 107: '+', 109: '-', 110: '.', 111: '/',
			186: ';', 187: '=', 188: ',', 189: '-', 190: '.',
			191: '/', 192: '`', 219: '[', 220: '\\', 221: ']', 222: "'",
			112: 'F1', 113: 'F2', 114: 'F3', 115: 'F4', 116: 'F5', 117: 'F6',
			118: 'F7', 119: 'F8', 120: 'F9', 121: 'F10', 122: 'F11', 123: 'F12'
		},
		isModifier: function(e) {
			return e.which > 15 && e.which < 19;
		},
		key: function(e) {
			return this.specKeys[e.which] || String.fromCharCode(e.which).toUpperCase();
		}
	};

	var root, initParams, panning, winW, winH, sX, sY;
	var noFit = {cur: false, real: false};
	var lastEvent = {};
	var progress = null;
	var cancelAction = false;
	var freeZoom = null;
	var dragSlide = [];
	var borderSize = 0;
	var MAXSIZE = 0x7fff;

	media.style.display = 'none';

	if (browser.opera) {
		if (mediaType() === 'img') {
			win.donotrun = media.error = true;
			media.naturalWidth = 0;

			media.onclick =
				win.ondragstart =
				win.onkeypress =
				win.onmousedown =
				win.onmouseup =
				win.onmousemove = null;
		}
		else {
			doc.body.replaceChild(media, media.parentNode);
		}
	}

	['class', 'style'].forEach(function(attr) {
		doc.documentElement.removeAttribute(attr);
		doc.body.removeAttribute(attr);
	});


	if (head = doc.querySelector('head')) {
		doc.documentElement.removeChild(head);
	}

	head = doc.createElement('head');

	if (cfg.favicon) {
		root = doc.createElement('link');
		root.rel = 'shortcut icon';
		root.href = cfg.favicon === '%url' && mediaType() === 'img'
			? win.location.href
			: cfg.favicon.replace('%url', encodeURIComponent(win.location.href));
	}

	if (root) {
		head.appendChild(root);
	}

	root = doc.createElement('style');
	root.appendChild(doc.createTextNode([
		'html, body{width: 100%; height: 100%; -webkit-user-select: none; -moz-user-select: none; -ms-user-select: none}',
		'body {cursor: default; margin:0; padding: 0; font: 12px "Trebuchet MS", sans-serif}',
		'body > img, body > video {margin:0 !important; background-clip: padding-box;' + browser['box-sizing_css'] + ': border-box; position: absolute}',
		'html.audio video{position: absolute; top: 0; right: 0; bottom: 0; left: 0; margin: auto !important;' +
			'min-width: 300px; width: 50% !important; height: 30px !important; background: transparent; box-shadow: none}',
		'h2 {margin: 0; padding-top: 15%; text-align: center; text-shadow: 1px 1px 15px #000; color: #fff; font: 700 30px "Trebuchet MS", sans-serif}',
		'.m_1 {max-width: 100%; max-height: 100%; width: auto; height: auto}',
		'.m_2 {width: 100%; height:auto} .m_3 {height: 100%; width: auto}',
		'a {text-decoration: none}',
		':root.error img, :root.error video {box-shadow: 0 0 10px red}',
		':root.error>body>h2 {text-shadow: 1px 1px 15px firebrick}',
		':focus {outline: 0}',
		':invalid {box-shadow: none}',
		'#menu {position: fixed; ' + browser['transition_css'] + ': opacity .25s, left .2s; width: 50px; height: 33.33%; top: 0; opacity: 0}',
		'ul {display: inline-block; text-align: center; margin: 0; padding: 5px 15px; list-style: none; font-size: 25px; font-weight: 700; background: rgba(0, 0, 0, .6); color: #fff}',
		'li {display: block; position: relative}',
		'li[data-cmd] {cursor: pointer}',
		'li[data-cmd=cycle], li[data-cmd=rotate] {position: relative}',
		'li[data-cmd=cycle]::after {content: "\u2195"; position: absolute; left: 25%}',
		'li[data-cmd=rotate]::after {content: "\u21BB"; position: absolute; left: 2px}',
		'li ul {position: absolute; left: 100%; top: 0; visibility: hidden; opacity: 0; ' + browser['transition_css'] + ': visibility .4s, opacity .2s .3s; text-align: left}',
		'li:hover ul {display: block; opacity: 1; visibility: visible}',
		'li ul>li {display: block !important; font-size: 15px; padding: 3px 0}',
		'.send_hosts li > a {color: #fff; background-size: 16px 16px; background-repeat: no-repeat; background-position: 0 3px; padding-left: 25px;}',
		'#menu>ul>li:hover, .send_hosts li > a:hover {color: silver}',
		'li ul>li {white-space: nowrap}',
		'.filters>ul {font-size: 15px} .filters>form {margin: 0; padding: 0}',
		'body.frames, body.frames {text-align: center; overflow-y: scroll !important} body.frames * {margin: 2px}',
		'#frames > canvas, #frames > img {display: none}',
		'#current_frame+output {display: inline-block; font-family: Conoslas, monospace; width: 65px; text-align: right}',
		'.back {font-size: 150%; color: black; text-decoration: none}',
		'#frames.showall > img, #frames.showall > canvas {display: inline-block !important}',
		(cfg.customCSS || 'body {background: silver; overflow: hidden} img,video {background: white; box-shadow: 0 0 5px grey}')
	].join('')));

	head.appendChild(root);
	doc.documentElement.insertBefore(head, doc.body);

	init = function() {
		if (!(media.naturalWidth || media.width || media.videoWidth
			|| mediaType() === 'audio' || initParams === null)) {
			if (initParams && ++initParams.loop >= initParams.maxLoop && progress) {
				clearInterval(progress);
			}

			return;
		}

		if (progress) {
			clearInterval(progress);
			progress = null;
		}

		root.classList.add(mediaType());

		if (mediaType() === 'img') {
			doc.body.replaceChild(media, doc.images[0]);
		}
		else if (mediaType() === 'video') {
			media.naturalWidth = media.videoWidth;
			media.naturalHeight = media.videoHeight;
		}

		[].forEach.call(doc.body.childNodes, function(node) {
			if (media !== node) {
				node.parentNode.removeChild(node);
			}
		});

		media.style.display = 'block';

		if (mediaType() === 'audio') {
			return;
		}

		head.appendChild(doc.createElement('style')).textContent =
			'body > img, body > video {max-width:' +
				Math.min(
					MAXSIZE,
					Math.floor(media.naturalWidth * MAXSIZE / media.naturalHeight)
				) +
				'px;max-height:' +
				Math.min(
					MAXSIZE,
					Math.floor(media.naturalHeight * MAXSIZE / media.naturalWidth)
				) + 'px}';

		if (!cfg.imgInfo) {
			cfg.imgInfo = false;
		}

		cfg.hiddenScrollbars = win.getComputedStyle(root).overflow === 'hidden'
										|| win.getComputedStyle(doc.body).overflow === 'hidden';

		var menu = doc.body.appendChild(doc.createElement('div'));
		menu.id = 'menu';

		if (win.getComputedStyle(menu).display === 'none') {
			doc.body.removeChild(menu);
			menu = null;
		}
		else {
			menu.style.cssText = '-webkit-filter: blur(1px); filter: blur(1px)';

			if (menu.style.webkitFilter || menu.style.filter) {
				media.filters = {};
				menu.onchange = function(e) {
					var t = e.target;
					var filterCSS = '';
					var filterName;

					if (t.value) {
						if (t.value === t.defaultValue) {
							delete media.filters[e.target.parentNode.textContent.trim()];
						}
						else {
							filterName = t.parentNode.textContent.trim();
							media.filters[filterName] = t.value + t.getAttribute('unit');
						}

						for (filterName in media.filters) {
							filterCSS += filterName + '(' + media.filters[filterName] + ') ';
						}
					}

					media.style.webkitFilter = filterCSS;
					media.style.filter = filterCSS;
				};
			}

			buildNodes(menu.appendChild(doc.createElement('ul')), [
				{tag: 'li', attrs: {'data-cmd': 'cycle'}, text: '\u2194'},
				{tag: 'li', attrs: {'data-cmd': 'zoom'}, nodes: [
					{tag: 'span', attrs: {style: 'visibility: hidden'}, text: '.'},
					{tag: 'span', attrs: {style: 'transform: rotate(-45deg); position: absolute; left: 20%; font-size: 125%'}, text: '\u26B2'}
				]},
				{tag: 'li', attrs: {'data-cmd': 'flip'}, text: '\u21CB'},
				{tag: 'li', attrs: {'data-cmd': 'rotate'}, text: '\u21BA'},
				media.filters ? {tag: 'li', attrs: {'class': 'filters'}, nodes: [
					'\u2261',
					{tag: 'form', nodes: [{tag: 'ul', nodes: [
						{tag: 'li', nodes: [{tag: 'input', attrs: {type: 'range', min: 0, max: 250, step: 10, value: 100, unit:'%'}}, ' brightness']},
						{tag: 'li', nodes: [{tag: 'input', attrs: {type: 'range', min: 0, max: 300, step: 25, value: 100, unit:'%'}}, ' contrast']},
						{tag: 'li', nodes: [{tag: 'input', attrs: {type: 'range', min: 0, max: 1000, step: 50, value: 100, unit:'%'}}, ' saturate']},
						{tag: 'li', nodes: [{tag: 'input', attrs: {type: 'range', min: 0, max: 100, step: 25, value: 0, unit:'%'}}, ' grayscale']},
						{tag: 'li', nodes: [{tag: 'input', attrs: {type: 'range', min: 0, max: 100, step: 100, value: 0, unit:'%'}}, ' invert']},
						{tag: 'li', nodes: [{tag: 'input', attrs: {type: 'range', min: 0, max: 100, step: 20, value: 0, unit:'%'}}, ' sepia']},
						{tag: 'li', nodes: [{tag: 'input', attrs: {type: 'range', min: 0, max: 360, step: 36, value: 0, unit:'deg'}}, ' hue-rotate']},
						{tag: 'li', nodes: [{tag: 'input', attrs: {type: 'range', min: 0, max: 20, step: 1, value: 0, unit:'px'}}, ' blur']}
					]}]},
				]} : '',
				{tag: 'li', attrs: {'data-cmd': 'reset'}, text: '\u2715'},
				(/^https?:$/.test(win.location.protocol) && cfg.sendTo.length
					? {tag: 'li', attrs: {'class': 'send_hosts'}, nodes: [
						'\u2197', {tag: 'ul', nodes: cfg.sendTo.map(function(item) {
							item = item.split('|');
							return {tag: 'li', nodes: [{tag: 'a', attrs: {href: item.slice(1).join('|')}, text: item[0]}]};
						})}
					]}
					: null),
				mediaType() === 'img' ? {tag: 'li', attrs: {'data-cmd': 'frames'}, text: '\u22EF'} : '',
				{tag: 'li', attrs: {'data-cmd': 'options'}, text: '\u2713'},
			]);

			menu.style.cssText = 'display: none; left: -' + menu.offsetWidth + 'px';

			menu.addEventListener('mousedown', function(e) {
				var t = e.target;

				if (t.href && t.href.indexOf('%', t.host.length) > -1) {
					t.href = t.href.replace(/%url/, encodeURIComponent(media.src));
				}

				pdsp(e, !!t.textContent);
			}, false);

			// Load favicons only when the menu item is hovered the first time
			if (/^https?:$/.test(win.location.protocol) && cfg.sendTo.length) {
				menu.querySelector('.send_hosts').onmouseover = function() {
					this.onmouseover = null;

					[].forEach.call(this.querySelectorAll('.send_hosts > ul > li > a'), function(a) {
						var url = a.protocol + '//' + (a.getAttribute('data-favicon') || (a.host + '/favicon.ico'));
						a.style.backgroundImage = 'url(' + url + ')';
					});
				};
			}

			var handleCommand = function(cmd, e) {
				if (e.button === 1) {
					return;
				}

				var p = e.type.indexOf('wheel') > -1 && (-e.wheelDelta || e.deltaY) > 0 || e.button === 2 ? 1 : 0;

				if (cmd === 'cycle') {
					media.cycle(!p);
				}
				else if (cmd === 'flip') {
					media.flip(media, p);
				}
				else if (cmd === 'rotate') {
					media.rotate(!p, e.ctrlKey);
				}
				else if (cmd === 'zoom') {
					pdsp(e);
					media.zoomToCenter({wheelDelta: p ? -1 : 1});
				}
				else if (cmd === 'reset' && e.button === 0) {
					media.reset();
				}
				else if (cmd === 'frames') {
					Port.listen(function(js) {
						Port.listen(null);
						Function('win', 'drawFullFrame', 'errorHandler', js['frames.js'])(win, e.button === 0, function(msg) {
							// Success
							if (msg !== null) {
								alert(msg);
								msg = menu.querySelector('li[data-cmd="frames"]');
								msg.removeAttribute('data-cmd');
								msg.style.opacity = '0.2';
								return;
							}

							if (browser.maxthon) {
								win.addEventListener('mousedown', function(e) {
									pdsp(e, false);
								}, true);
							}
						});
					});

					Port.send({cmd: 'load_frames.js'});
				}
				else if (cmd === 'options' && e.button === 0) {
					Port.send({cmd: 'open', url: 'options.html'});
				}
				else {
					p = null;
				}

				if (p !== null) {
					pdsp(e);
				}
			};

			menu.addEventListener(browser['wheel'], function(e) {
				var t = e.target;

				if (t.nodeType === 3) {
					t = t.parentNode;
				}

				if (t.type === 'range') {
					t.value = Math.max(
						t.getAttribute('min'),
						Math.min(
							parseInt(t.value, 10) + t.getAttribute('step') * ((-e.wheelDelta || e.deltaY) > 0 ? -1 : 1),
							t.getAttribute('max')
						)
					);
					menu.onchange(e);
					pdsp(e);
				}
				else if (!/reset|frames|options/.test(t.getAttribute('data-cmd')
					|| (t = t.parentNode) && t.getAttribute('data-cmd'))) {
					handleCommand(t.getAttribute('data-cmd'), e);
				}
			}, false);

			menu.onclick = function(e) {
				var t = e.target;

				if (t.hasAttribute('data-cmd')
					|| (t = t.parentNode) && t.hasAttribute('data-cmd')) {
					handleCommand(t.getAttribute('data-cmd'), e);
				}
			};

			menu.oncontextmenu = function(e) {
				var t = e.target;

				if (t.type === 'range') {
					t.value = t.defaultValue;
					menu.onchange(e);
					pdsp(e);
				}
				else if (t.hasAttribute('data-cmd')
					|| (t = t.parentNode) && t.hasAttribute('data-cmd')) {
					handleCommand(t.getAttribute('data-cmd'), e);
				}
			};

			menu.addEventListener(browser['transitionend'], function(e) {
				if (e.propertyName === 'left' && this.style.left[0] === '-') {
					menu.style.display = 'none';
				}
			}, false);

			if (win.Node.prototype && !win.Node.prototype.contains) {
				win.Node.prototype.contains = function(n) {
					return n instanceof Node && (this === n || !!(this.compareDocumentPosition(n) & 16));
				};
			}

			menu.onmouseover = function() {
				if (!menu.mtimer) {
					return;
				}

				clearTimeout(menu.mtimer);
				menu.mtimer = null;
				menu.style.left = '0';
				menu.style.opacity = '1';
			};

			menu.onmouseout = function(e) {
				if (this.contains(e.relatedTarget)) {
					return;
				}

				doc.addEventListener('mousemove', menuTrigger, false);
				menu.mtimer = setTimeout(function() {
					menu.style.left = '-' + menu.offsetWidth + 'px';
					menu.style.opacity = '0';
					menu.mtimer = null;
				}, 800);
			};

			var menuTrigger = function(e) {
				if (panning || freeZoom) {
					return;
				}

				if (e.clientX > 40 || e.clientY > win.innerHeight / 3) {
					return;
				}

				if (menu.style.display === 'block') {
					return;
				}

				menu.style.display = 'block';

				setTimeout(function() {
					menu.style.left = '0';
					menu.style.opacity = '1';
				}, 50);

				doc.removeEventListener('mousemove', menuTrigger, false);
			};

			// Safari showed the menu even if the cursor wasn't at the edge
			setTimeout(function() {
				doc.addEventListener('mousemove', menuTrigger, false);
			}, 500);
		}

		var afterCalcTimeoutCallback = function() {
			// Gecko and WebKit/Blink render scrollbars for a split of a second,
			// even if the image fits into the view-port
			winH = doc.compatMode[0] === 'B' ? doc.body : root;
			winW = winH.clientWidth;
			winH = winH.clientHeight;

			media.calcFit();
			media.setInfo();
			media.setCur();
		};

		var afterCalc = function() {
			if (media.calcFit) {
				winH = doc.compatMode[0] === 'B' ? doc.body : root;
				winW = winH.clientWidth;
				winH = winH.clientHeight;

				media.setPos();

				setTimeout(afterCalcTimeoutCallback, 0xf);
			}
		};

		media.resize = function(m, w) {
			if (m === -1 || (m === -1 && freeZoom)) {
				if (w) {
					this.style.width = w;
				}
			}

			if (m === void 0) {
				m = 1;
			}

			if (m !== -1 && this.mode === -1) {
				this.style.width = this.style.height = '';
			}

			this.mode = m;
			this.className = 'm_' + m;

			if (this.naturalWidth) {
				afterCalc();
			}
		};

		media.setPos = function() {
			var s = this.style;

			if (cfg.center) {
				s.top = Math.max(0, (winH - this.offsetHeight) / 2) + 'px';
				s.left = Math.max(0, (winW - this.offsetWidth) / 2) + 'px';
			}
			else {
				s.top = s.left = '0';
			}

			this.box = this.getBoundingClientRect();

			if (this.box.left < 0) {
				s.left = (parseInt(s.left, 10) - this.box.left - win.pageXOffset) + 'px';
			}

			if (this.box.top < 0) {
				s.top = (parseInt(s.top, 10) - this.box.top - win.pageYOffset) + 'px';
			}
		};

		media.setInfo = function() {
			if (!cfg.imgInfo || !(media.naturalWidth && media.naturalHeight)) {
				return;
			}

			doc.title = cfg.imgInfo.replace(/%(o?[wh]|url|name|ratio|perc)/g, function(a, a1) {
				switch (a1) {
					case 'w': return media.clientWidth;
					case 'h': return media.clientHeight;
					case 'ow': return media.naturalWidth;
					case 'oh': return media.naturalHeight;
					case 'url': return win.location.href;
					case 'name': return media.alt;
					case 'ratio':
						return Math.round(media.clientWidth / media.clientHeight * 100) / 100;
					case 'perc':
						return ~~(0.5 + 1e4 * (media.clientWidth / media.naturalWidth)) / 100;
				}
			});
		};

		media.calcFit = function() {
			var box = this.box || this.getBoundingClientRect();

			noFit = {
				cur: box.width <= winW && box.height <= winH,
				real: this.naturalWidth <= winW && this.naturalHeight <= winH
			};
		};

		media.wheelZoom = function(e) {
			stopScroll();
			pdsp(e);

			var w = media.offsetWidth;
			var h = media.offsetHeight;

			if ((-e.wheelDelta || e.deltaY) > 0) {
				media.resize(-1, Math.max(1, w * 0.75) + 'px');
			}
			else {
				var width = w * (4 / 3);

				media.resize(-1, (width > 10 ? width : width + 3) + 'px');
			}

			if (e.keypress || e.target.nodeName.toUpperCase() === 'IMG') {
				win.scrollTo(
					(e.offsetX || e.layerX || 0) * media.offsetWidth / w - e.clientX + borderSize,
					(e.offsetY || e.layerY || 0) * media.offsetHeight / h - e.clientY + borderSize
				);
			}
		};

		media.zoomToCenter = function(e) {
			pdsp(e);
			this.wheelZoom({
				keypress: true,
				wheelDelta: e.wheelDelta,
				clientX: winW / 2,
				clientY: winH / 2,
				offsetX: win.pageXOffset + media.offsetLeft + winW / 2,
				offsetY: win.pageYOffset + media.offsetTop + winH / 2
			});
		};

		media.cycle = function(back) {
			var mode = (this.mode === 1 ? 0 : this.mode) - (back ? 1 : -1);

			if (mode === 1) {
				mode = mode - (back ? 1 : -1);
			}
			else if (mode < 0) {
				mode = 3;
			}
			else if (mode > 3) {
				mode = 0;
			}

			this.resize(mode);
		};

		media.reset = function() {
			if (freeZoom) {
				doc.removeEventListener('mousemove', drawMask, false);
				doc.body.removeChild(media.mask);

				freeZoom = null;
				cancelAction = false;
				return;
			}

			// Firefox document readyState stucks in 'interactive' for videos,
			// but videos doesn't need to be loaded anyway
			if (mediaType() === 'img' && doc.readyState !== 'complete') {
				return;
			}

			var filters = doc.body.querySelector('#menu li.filters > form');

			if (filters) {
				this.style.filter = this.style.webkitFilter = '';
				media.filters = {};
				filters.reset();
			}

			delete this.curdeg;
			delete this.scale;
			this.style[browser['transform']] =
					this.style.width = this.style.height = '';
			this.resize(0);
		};

		media.flip = function(el, ori) {
			if (!this.scale) {
				this.scale = {h: 1, v: 1};
			}

			this.scale[ori ? 'h' : 'v'] *= -1;
			ori = this.scale.h !== 1 || this.scale.v !== 1 ?
				'scale(' + this.scale.h + ',' + this.scale.v + ')' :
				'';

			if (this.curdeg) {
				ori += ' rotate(' + this.curdeg + 'deg)';
			}

			this.style[browser['transform']] = ori;
		};

		media.rotate = function(deg, fine) {
			var rot;

			if (!this.curdeg) {
				this.curdeg = 0;
			}

			if (deg) {
				this.curdeg += (fine ? 10 : 90);
			}
			else {
				this.curdeg -= (fine ? 10 : 90);
			}

			win.status = this.curdeg + '°';

			rot = 'rotate(' + this.curdeg + 'deg)';

			if (this.scale) {
				rot += ' scale(' + this.scale.h + ', ' + this.scale.v + ')';
			}

			this.style[browser['transform']] = rot;
			this.setPos();
		};

		if (cfg.wheelZoom) {
			doc.addEventListener(browser['wheel'], media.wheelZoom, false);
		}
		else {
			doc.addEventListener(browser['wheel'], function(e) {
				if (media.clientWidth > winW || media.clientHeight > winH) {
					stopScroll();

					var x = 0;
					var y = ((-e.wheelDelta || e.deltaY) > 0 ? winH : -winH) / 5;

					if (media.clientWidth <= winW && media.clientHeight > winH) {
						if (!cfg.hiddenScrollbars) {
							return;
						}
					}
					else if (media.clientHeight <= winH && media.clientWidth > winW
						|| e.clientX < winW / 2 && e.clientY > winH - 100) {
						x = (y < 0 ? -winW : winW) / 5;
						y = 0;
					}
					else if (!cfg.hiddenScrollbars) {
						return;
					}

					win.scrollBy(x, y);
					pdsp(e);
				}
			}, false);
		}

		media.setCur = function() {
			if (this.clientWidth > winW || this.clientHeight > winH) {
				this.style.cursor = 'move';
			}
			else if (
				((winH >= this.clientHeight && winW >= this.clientWidth)
					&& (winH < this.naturalHeight || winW < this.naturalWidth))
				|| (noFit.cur && noFit.real
					&& (this.clientHeight < this.naturalHeight
						|| this.clientWidth < this.naturalWidth))) {
				this.style.cursor = browser['zoom-in'];
			}
			else {
				this.style.cursor = 'default';
			}
		};

		win.addEventListener('resize', afterCalc, false);

		doc.addEventListener('contextmenu', function(e) {
			doc.removeEventListener('mousemove', onmove, true);

			if (progress) {
				clearTimeout(progress);
				progress = null;
			}

			if (cancelAction) {
				e.preventDefault();
			}
		}, false);

		var lastMoveX, lastMoveY;

		var onmove = function(e) {
			lastMoveX = e.clientX;
			lastMoveY = e.clientY;

			if (lastMoveX === lastEvent.clientX && lastMoveY === lastEvent.clientY) {
				return;
			}
			else if (progress) {
				clearTimeout(progress);
				progress = lastEvent.button = null;
			}

			if (sX === true || noFit.cur) {
				doc.removeEventListener('mousemove', onmove, true);
				return;
			}

			if (!panning) {
				panning = win.requestAnimationFrame(onmoveframe);
			}

			if (dragSlide.length !== 3) {
				dragSlide = [
					[lastMoveX, lastMoveY],
					[lastMoveX, lastMoveY],
					[lastMoveX, lastMoveY]
				];
				return;
			}

			// Opera fires move event before mouseup
			if (dragSlide[2][0] === lastMoveX && dragSlide[2][1] === lastMoveY) {
				return;
			}

			dragSlide[0] = [dragSlide[1][0], dragSlide[1][1]];
			dragSlide[1] = [dragSlide[2][0], dragSlide[2][1]];
			dragSlide[2] = [lastMoveX, lastMoveY];

			media.dragSlideTime = Date.now();
		};

		var pan = function() {
			win.scrollBy(sX - lastMoveX, sY - lastMoveY);

			sX = lastMoveX;
			sY = lastMoveY;
		};

		var onmoveframe = function() {
			pan();
			panning = null;
		};

		var startScroll = function() {
			progress = setInterval(function() {
				win.scrollBy(dragSlide[0], dragSlide[1]);

				if (dragSlide[0]) {
					if (Math.abs(dragSlide[0]) < 1) {
						dragSlide[0] = 0;
					}

					dragSlide[0] /= 1.11;
				}

				if (dragSlide[1]) {
					if (Math.abs(dragSlide[1]) < 1) {
						dragSlide[1] = 0;
					}

					dragSlide[1] /= 1.11;
				}

				var
					atRight = root.scrollWidth - win.pageXOffset === winW,
					atBottom = root.scrollHeight - win.pageYOffset === winH;

				if (!(dragSlide[0] && dragSlide[1])
						|| (!win.pageYOffset && !win.pageXOffset)
						|| (!win.pageYOffset && atRight)
						|| (atBottom && !win.pageXOffset) || (atBottom && atRight)) {
					stopScroll();
				}
			}, 25);
		};

		var stopScroll = function(e) {
			if (dragSlide.length) {
				clearInterval(progress);
				progress = null;
				cancelAction = false;
				dragSlide.length = 0;
				media.dragSlideTime = false;
			}

			if (e) {
				doc.removeEventListener(browser['wheel'], stopScroll, false);
			}
		};

		var drawMask = function(e) {
			if (++freeZoom.counter % 3) {
				return;
			}

			var x = e.clientX - freeZoom.left;
			var y = e.clientY - freeZoom.top;

			if (e.ctrlKey) {
				var rx = freeZoom.prevX ? x - freeZoom.prevX : 0;
				var ry = freeZoom.prevY ? y - freeZoom.prevY : 0;

				freeZoom.prevX = x;
				freeZoom.prevY = y;

				freeZoom.x += rx;
				freeZoom.y += ry;
				freeZoom.X += rx;
				freeZoom.Y += ry;
			}
			else {
				if (freeZoom.prevX) {
					delete freeZoom.prevX;
					delete freeZoom.prevY;
				}

				freeZoom.w = Math.abs(freeZoom.X - x);
				freeZoom.h = Math.abs(freeZoom.Y - y);
				freeZoom.x = freeZoom.X < x ? freeZoom.X : freeZoom.X - freeZoom.w;
				freeZoom.y = freeZoom.Y < y ? freeZoom.Y : freeZoom.Y - freeZoom.h;
			}

			media.mctx.clearRect(0, 0, media.mask.width, media.mask.height);
			media.mctx.fillRect(0, 0, media.mask.width, media.mask.height);
			media.mctx.clearRect(freeZoom.x, freeZoom.y, freeZoom.w, freeZoom.h);
		};

		var longpressHandler = function() {
			cancelAction = true;
			progress = null;

			switch (cfg[lastEvent.button === 2 ? 'lpRight' : 'lpLeft']) {
				// Fit to width/height
				case 1:
					var x, y;

					if (media.mode === 2 || media.clientWidth === doc.body.clientWidth
						|| (doc.body.clientHeight > media.clientHeight && winW < doc.body.clientWidth)) {
						x = media.clientWidth;
						media.resize(3);
						x = lastEvent.layerX * media.clientWidth / x - winW / 2;
						y = 0;
					}
					else {
						y = media.clientHeight;
						media.resize(2);
						x = 0;
						y = lastEvent.layerY * media.clientHeight / y - winH / 2;
					}

					win.scrollTo(x, y);
				break;
			}
		};

		media.addEventListener('mousedown', function(e) {
			pdsp(e, false);

			if (e.button === 1 || e.ctrlKey || e.altKey
				|| !e.shiftKey && mediaType() === 'video'
					&& (e.offsetY || e.layerY || 0) > Math.min(media.clientHeight - 40, media.clientHeight / 2)) {
				return;
			}

			if ((e.button === 0 && media.mode < 4) || e.button === 2) {
				if (e.button === 2) {
					sX = true;
				}
				else {
					if (!e.shiftKey && noFit.cur && media.clientWidth >= 60
						&& media.clientWidth - (e.offsetX || e.layerX || 0) <= 30) {
						lastEvent.clientX = e.clientX;
						lastEvent.clientY = e.clientY;
						return;
					}

					e.preventDefault();

					win.focus();
					sX = e.clientX;
					sY = e.clientY;
				}

				if (!e.shiftKey) {
					doc.addEventListener('mousemove', onmove, true);
				}
			}

			if (!cfg.lpDelay) {
				return;
			}

			// For fine move and free zoom
			if (e.shiftKey) {
				cancelAction = true;

				if (e.button === 0) {
					freeZoom = {
						counter: 0,
						left: parseInt(media.style.left, 10),
						top: parseInt(media.style.top, 10)
					};

					freeZoom.X = e.clientX - freeZoom.left;
					freeZoom.Y = e.clientY - freeZoom.top;

					if (!media.mask) {
						media.mask = doc.createElement('canvas');
						media.mask.className = 'mask';
						media.mask.style.cssText = 'display: block; position: fixed; left: 0; top: 0;';
						media.mctx = media.mask.getContext('2d');
					}

					doc.addEventListener('mousemove', drawMask, false);

					var h = media.curdeg && Math.sin(media.curdeg) ? true : false;
					var w = h ? media.offsetHeight : media.offsetWidth;
					h = h ? media.offsetWidth : media.offsetHeight;
					media.mask.width = Math.min(winW, w);
					media.mask.height = Math.min(winH, h);

					media.mask.style.left = media.style.left;
					media.mask.style.top = media.style.top;

					media.mctx.clearRect(0, 0, media.mask.width, media.mask.height);
					doc.body.appendChild(media.mask);

					w = win.getComputedStyle(media.mask).color;
					media.mctx.fillStyle = !w || w === 'rgb(0, 0, 0)' ? 'rgba(0,0,0,.4)' : w;
				}
			}

			// Is dragSlideing
			if (progress) {
				if (stopScroll) {
					stopScroll();
					cancelAction = e.button === 0;
				}

				if (e.button === 0) {
					return;
				}
			}

			if (e.shiftKey || (e.button === 0 && !cfg.lpLeft)
				|| (e.button === 2 && !cfg.lpRight)) {
				return;
			}

			lastEvent.clientX = e.clientX;
			lastEvent.clientY = e.clientY;
			lastEvent.layerX = e.offsetX || e.layerX || 0;
			lastEvent.layerY = e.offsetY || e.layerY || 0;
			lastEvent.button = e.button;

			progress = setTimeout(longpressHandler, cfg.lpDelay);
		}, true);

		doc.addEventListener('mouseup', function(e) {
			if (e.button !== 0) {
				return;
			}

			var x, y;

			if (media.mode < 4) {
				doc.removeEventListener('mousemove', onmove, true);

				if (freeZoom) {
					doc.removeEventListener('mousemove', drawMask, false);
					doc.body.removeChild(media.mask);

					var w = Math.min(
						media.clientWidth,
						freeZoom.w + Math.min(freeZoom.x, 0)
							+ (freeZoom.x + freeZoom.w > media.mask.width ?
									media.mask.width - freeZoom.x - freeZoom.w : 0)
					);

					var h = Math.min(
						media.clientHeight,
						freeZoom.h + Math.min(freeZoom.y, 0)
							+ (freeZoom.y + freeZoom.h > media.mask.height ?
									media.mask.height - freeZoom.y - freeZoom.h : 0)
					);

					x = Math.max(0, freeZoom.x);
					y = Math.max(0, freeZoom.y);

					freeZoom = null;
					cancelAction = false;

					if (!w || !h || x >= media.clientWidth || y >= media.clientHeight) {
						return;
					}

					var nimgw, nimgh;
					var scrollbars = (win.innerWidth - winW) || (win.innerHeight - winH);
					var fitW = winW < w * winH / h;

					fitW = e.ctrlKey ? !fitW : fitW;

					if (fitW) {
						nimgw = media.clientWidth * winW / w;
						nimgh = nimgw * media.clientHeight / media.clientWidth;
					}
					else {
						nimgh = media.clientHeight * winH / h;
						nimgw = nimgh * media.clientWidth / media.clientHeight;
					}

					if (nimgw > MAXSIZE) {
						nimgw = MAXSIZE;
						nimgh = media.clientHeight * MAXSIZE / media.clientWidth;
					}

					if (nimgh > MAXSIZE) {
						nimgw = media.clientWidth * MAXSIZE / media.clientHeight;
						nimgh = MAXSIZE;
					}

					var cx = (win.pageXOffset + x + w / 2) * nimgw /
										media.clientWidth - winW / 2 - scrollbars;
					var cy = (win.pageYOffset + y + h / 2) * nimgh /
										media.clientHeight - winH / 2 - scrollbars;

					media.resize(-1, nimgw + 'px');

					win.scrollTo(cx, cy);
					return;
				}
				else if (dragSlide.length === 3) {
					x = dragSlide[0][0] - dragSlide[2][0];
					y = dragSlide[0][1] - dragSlide[2][1];

					if ((Date.now() - media.dragSlideTime) > 100) {
						cancelAction = false;
						dragSlide.length = 0;
						return;
					}

					if (x || y) {
						dragSlide.length = 2;
						dragSlide[0] = x;
						dragSlide[1] = y;
						startScroll();
						doc.addEventListener(browser['wheel'], stopScroll, false);
						return;
					}
				}
			}

			if (progress) {
				clearTimeout(progress);
				progress = null;
			}

			if (cancelAction) {
				cancelAction = false;
				return;
			}

			if (e.shiftKey || e.ctrlKey || e.altKey
				|| e.target !== media || lastEvent.button === null
				|| e.clientX !== lastEvent.clientX || e.clientY !== lastEvent.clientY) {
				return;
			}

			if (media.mode < 2 && noFit.real) {
				if (media.mode === -1) {
					media.resize(0);
				}
				else if (winW / winH < media.naturalWidth / media.naturalHeight
					|| winH === media.offsetHeight && winW > media.offsetWidth) {
					media.resize(2);
				}
				else {
					media.resize(3);
				}
			}
			else if (media.mode === 1 || noFit.cur) {
				if (media.naturalWidth === media.clientWidth
					&& media.naturalHeight === media.clientHeight) {
					return;
				}

				x = e.offsetX || e.layerX || 0;
				y = e.offsetY || e.layerY || 0;

				if (media.scale) {
					if (media.scale.h === -1) {
						x = media.clientWidth - x;
					}

					if (media.scale.v === -1) {
						y = media.clientHeight - y;
					}
				}

				x = x * media.naturalWidth / media.clientWidth - winW / 2;
				y = y * media.naturalHeight / media.clientHeight - winH / 2;

				media.resize(0);
				win.scrollTo(x, y);
			}
			else {
				media.resize(1);
			}
		}, true);

		doc.addEventListener('keydown', function(e) {
			if (stopScroll) {
				stopScroll();
			}

			if (shortcut.isModifier(e)) {
				return;
			}

			var key = shortcut.key(e);

			if (e.ctrlKey || (freeZoom && key !== 'Esc')) {
				return;
			}

			if (key === '+' || key === '-') {
				e.wheelDelta = key === '+' ? 1 : -1;
				media.zoomToCenter(e);
				return false;
			}

			var x, y, z;

			switch (key) {
				case 'Esc':
					media.reset();
				break;

				case 'PgUp':
					x = e.shiftKey ? -winW / 2 : 0;
					y = e.shiftKey ? 0 : -winH / 2;
				break;

				case 'PgDn':
					x = e.shiftKey ? winW / 2 : 0;
					y = e.shiftKey ? 0 : winH / 2;
				break;

				case 'End':
					x = e.shiftKey ? root.scrollWidth : win.pageXOffset;
					y = e.shiftKey ? win.pageYOffset : root.scrollHeight;
					z = true;
				break;

				case 'Home':
					x = e.shiftKey ? 0 : win.pageXOffset;
					y = e.shiftKey ? win.pageYOffset : 0;
					z = true;
				break;

				default: x = true;
			}

			if (x !== true) {
				if (x !== void 0) {
					win[z ? 'scrollTo' : 'scrollBy'](x, y);
					e.preventDefault();
				}

				return;
			}

			switch (key) {
				case cfg.key_mOrig: media.resize(0); break;
				case cfg.key_mFit: media.resize(1); break;
				case cfg.key_mFitW: media.resize(2); break;
				case cfg.key_mFitH: media.resize(3); break;
				case cfg.key_cycle: media.cycle(e.shiftKey); break;
				case cfg.key_rotL: media.rotate(false, e.shiftKey); break;
				case cfg.key_rotR: media.rotate(true, e.shiftKey); break;
				case cfg.key_flipH: media.flip(media, 0); break;
				case cfg.key_flipV: media.flip(media, 1); break;
				default: x = true;
			}

			if (mediaType() === 'video') {
				x = null;

				switch (key) {
					case 'Space': media.togglePlay(); break;
					case 'Up': case 'Down':
						key = media.volume + (key === 'Up' ? 0.1 : -0.1);
						// Must check the range manually, otherwise browsers will throw an error
						media.volume = key < 0 ? 0 : (key > 1 ? 1 : key);
						break;
					default: x = true;
				}
			}

			if (x !== true) {
				pdsp(e);
			}
		}, true);

		if (mediaType() === 'video') {
			media.addEventListener('click', function(e) {
				var top = (e.offsetY || e.layerY || 0);

				if (top > media.clientHeight - 40) {
					return;
				}

				if (top > media.clientHeight / 2) {
					media.togglePlay();
				}

				pdsp(e);
			});

			media.addEventListener('dblclick', function(e) {
				if ((e.offsetY || e.layerY || 0) < media.clientHeight / 2) {
					pdsp(e);
				}
			});
		}

		media.mode = cfg.mode;
		media.calcFit();
		progress = [];

		if (cfg.minUpscale) {
			if (media.naturalWidth >= winW * cfg.minUpscale / 100) {
				progress[0] = 1;
			}

			if (media.naturalHeight >= winH * cfg.minUpscale / 100) {
				progress[1] = 1;
			}
		}

		if (media.mode === 4) {
			media.mode = media.naturalWidth / media.naturalHeight > winW / winH ? 3 : 2;
		}
		else if ((media.mode === 1 || (media.mode === 0 && noFit.real))
			&& progress.length) {
			media.mode = media.naturalWidth / media.naturalHeight > winW / winH ? 2 : 3;
		}

		if ((media.mode === 2 && media.naturalWidth < winW && !progress[0])
			|| (media.mode === 3 && media.naturalHeight < winH && !progress[1])) {
			media.mode = 0;
		}

		progress = null;
		media.resize(media.mode);
	};

	if (mediaType() === 'video') {
		root = {};
		cfg.vidattrs.split(/\s+/).forEach(function(attr) {
			attr = attr.split('=');

			if (attr[0] === 'volume') {
				root[attr[0]] = Math.min(100, Math.max(0, parseInt(attr[1]) / 100));
			}
			else {
				root[attr[0]] = true;
			}
		});

		['autoplay', 'loop', 'controls', 'muted', 'volume'].forEach(function(attr) {
			if (attr === 'volume') {
				media[attr] = root[attr] === void 0 ? 1 : root[attr];
			}
			else {
				media[attr] = root[attr] || false;
			}
		});

		media.togglePlay = function() {
			if (this.paused || Math.abs(this.duration - this.currentTime) < 0.01) {
				this.play();
			}
			else {
				this.pause();
			}
		};

		media.onloadedmetadata = function() {
			this.onloadedmetadata = null;

			if (!this.videoHeight) {
				mediaType('audio');

				document.onkeydown = function(e) {
					var key = shortcut.key(e);

					if (key === 'Space') {
						media.togglePlay();
					}
					else if (key === 'Up' || key === 'Down') {
						key = media.volume + (key === 'Up' ? 0.1 : -0.1);
						media.volume = key < 0 ? 0 : (key > 1 ? 1 : key);
					}
					else {
						return;
					}

					pdsp(e);
				};

				media.ondblclick = function(e) {
					pdsp(e);
				};

				if (browser.chrome || browser.maxthon || browser.safari) {
					media.onplaying = function() {
						this.poster = 'data:image/gif;base64,R0lGODlhAQABAAAAACw=';
					};
				}
			}

			if (!media.autoplay) {
				return;
			}

			setTimeout(function() {
				if (media.currentTime) {
					media.currentTime = 0;
					media.play();
				}
			}, 100);
		};
	}
	else {
		media = doc.createElement('img');
		media.src = win.location.href;
	}

	root = doc.documentElement;

	media.onerror = function() {
		// Opera fires an error event on local video files when they're close to the end
		// Firefox did it a few times too, however there was nothing wrong with the video
		if (media.currentTime > 100 && win.location.protocol === 'file:') {
			return;
		}

		clearInterval(progress);
		root.classList.add('error');
	};

	if (win.location.protocol === 'data:') {
		media.alt = mediaType() + ' (data:)';

		if (!cfg.imgInfo) {
			doc.title = media.alt;
		}
	}
	else {
		media.alt = (win.location.href.replace(/#.*/, '')
			.match(/(?:[^\/]+)?$/)[0] || mediaType())
			.split('?')[0];

		try {
			// Some Unicode characters caused problems for decodeURIComponent
			media.alt = decodeURIComponent(media.alt);
		} catch (ex) {}
	}

	if (media.naturalWidth || media.videoWidth || mediaType() === 'audio') {
		initParams = null;
		// Firefox and Safari won't position the image without setTimeout
		setTimeout(init, 0xf);
	}
	else {
		initParams = {loop: 0, wait: 150, maxLoop: 900};
		progress = setInterval(init, initParams.wait);

		setTimeout(function() {
			if (root.classList.contains(mediaType())) {
				return;
			}

			// Firefox won't show the image if a site uses CSP settings,
			// so we should use the default image
			if ((browser.firefox || browser.maxthon)
				&& (mediaType() === 'img'
					? !media.naturalWidth && doc.images[0].naturalWidth
					: !media.videoWidth && doc.body.querySelector('video').videoWidth)) {
				doc.images[0].style.display = 'block';
				clearInterval(progress);
				return;
			}

			media.style.display = 'none';
			var h2 = doc.body.appendChild(doc.createElement('h2'));
			h2.textContent = media.alt;
			h2.onclick = function() {
				doc.body.removeChild(this);
				media.style.display = 'block';
			};
		}, 1000);
	}
};

(function() {
	if (browser.safari && location.protocol === 'safari-extension:') {
		init = null;
		return;
	}

	if (browser.opera || browser.firefox) {
		if (mediaType()) {
			Port.listen(function(e) {
				Port.listen(null);
				init(window, document, e);
			});
			Port.send({cmd: 'hello'});
		}
		else {
			init = null;
		}
	}
	else {
		Port.listen(function(e) {
			Port.listen(null);

			if (!mediaType()) {
				init = null;
				return;
			}

			init(window, document, e);
		});

		Port.send({cmd: 'hello'});
	}
})();
